""" mind.py
 Copyright (C) 1998 Aloril
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

from copy import copy
from types import *

from task import task,task_info
from thing import thing
from event import *
from know import know
from mem_map import mem_map
from debug import debug
import wtime
import dictlist, const, geometry, mgoal

class memory:
    def __init__(self):
        self.events=[]
        self.places={}
    def destroy(self):
        self.events=None
        self.places=None
    def remember_place(self, place):
        self.places[place.get_xyz()]=place
    def recall_place(self, xyz, radius=0):
        if not radius:
            try:
                [(0.0,self.places[xyz])]
            except KeyError:
                return []
        res=[]
        for p in self.places.keys():
            d=geometry.distance(p,xyz)
            if d<=radius:
                res.append((d,self.places[p]))
        res.sort()
        return res
    def remember_event(self, event):
        "add new memory with age"
        self.events.append([event,1.0])
    def recall_event(self, event, cmp):
        "return list of memories with same command"
        found=[]
        for (e,age) in self.events:
            if apply(cmp,(event,e)): found.append(e)
        return found
    def forget(self):
        "age memories and forgot too old ones"
        for m in self.events:
            m[1]=m[1]-0.1
        #remove forgotten things
        self.events=filter(lambda m:m[1]>const.fzero,self.events)

class job:
    def __init__(self, main_task=None):
        self.main_task=main_task
        self.main_task_info=task_info(main_task)
        self.task_list=[]
    def __str__(self):
        return "<Job: Main task: "+str(self.main_task)+\
               "\nTask list: "+str(self.task_list)+">\n"
    def __repr__(self):
        return "<Job: "+self.main_task.__doc__+">"
    def __cmp__(self,other):
        importance=wtime.cmp_importance(self.main_task.importance,\
                                        other.main_task.importance)
        if importance!=0: return importance
        return self.no-other.no
    def destroy(self):
        self.main_task=None
        self.main_task_info=None
    def list_required_things(self):
        return self.main_task.objects
    def do(self):
        return self.main_task.do(self.main_task_info)

class eye(thing):
    def __init__(self, mind):
        thing.__init__(self,mind.name)
        self.mind=mind
    def destroy(self):
        self.mind=None
    def move_event(self,e):
        m=self.mind
        target=e.loc
        w=e.what
        if e.loc==m: m.add_thing(w)
        #if its part of inventory, use its xyz
        if hasattr(target,"xyz"):
            w.xyz=None
            w.place=target
        else:
            w.xyz=target
            w.place=None
        if w==m and not m.place:
            m.place=m.get_rknowledge("xyz",w.xyz)
    def make_event(self,e):
        m=self.mind
        if e.source==m:
            m.add_thing(e.what)
    def change_event(self,e):
        t=e.what
        t.status=t.status+e.amount
    def fire_event(self,e):
        m=self.mind
        if e.loc!=m.place:
            return [event("move",what=m, loc=e.loc)]
        return [event("extinguish",what=e.what, target=e.what)]
    def destroy_event(self,e):
        return self.mind.destroy_event(e)

class mind(thing):
    "minds: (=npc)"
    def __init__(self, name, desc='thinking thing', place='home', tasks=[],
                 status=1.0, age=0.0, xyz=0.0,
                 personal_knowledge=None,
                 local_knowledge=None,
                 common_knowledge=None,
                 common_goals=[]):
        thing.__init__(self,name,desc,place,status,age,xyz)
        self.eyes=eye(self)
        self.tasks=copy(tasks)
        self.current_jobs=[]
        if personal_knowledge: self.personal_knowledge=personal_knowledge
        else: self.personal_knowledge=know()
        if local_knowledge: self.local_knowledge=local_knowledge
        else: self.local_knowledge=know()
        if common_knowledge: self.common_knowledge=common_knowledge
        else: self.common_knowledge=know()
        if xyz: self.xyz=xyz
        else: self.xyz=self.get_knowledge("xyz",place)
        self.mem=memory()
        self.map=mem_map()
        self.things={}
        self.reverse_knowledge()
        self.common_goals=common_goals[:]
        self.debug=debug(self.name+".mind.log")
    def __str__(self):
        s="<Mind: "+thing.str(self)+"\n"
        s=s+"current jobs: "+`self.current_jobs`+"\n"
        s=s+"tasks: "+`self.tasks`+"\n"
        s=s+"things: "+`self.things`+">\n"
        return s+">\n"
    def destroy(self):
        thing.destroy(self)
        self.things=None
        self.mem.destroy()
        for j in self.current_jobs:
            j.destroy()
        self.current_jobs=None
        self.body=None
        self.eyes.destroy()
        self.eyes=None
    def reverse_knowledge(self):
        self.rknow=know()
        for k in (self.personal_knowledge,self.local_knowledge,self.common_knowledge):
            for (k,v) in k.xyz.items():
                if not self.rknow.xyz.get(v):
                    self.rknow.add("xyz",v,k)
    def get_rknowledge(self, what, key):
        d=getattr(self.rknow,what)
        return d.get(key)
    def get_knowledge(self, what, key):
        for k in (self.personal_knowledge,self.local_knowledge,self.common_knowledge):
            d=getattr(k,what)
            try: return d[key]
            except KeyError: pass
        return None
    def add_knowledge(self,what,key,value):
        self.personal_knowledge.add(what,key,value)
        if what=="xyz":
            self.rknow.add("xyz",value,key)
    def add_task(self, task): self.tasks.append(task)
    def add_thing(self,thing): 
        if not thing.xyz:
            thing.xyz=self.get_knowledge("xyz",thing.place)
        dictlist.add_value(self.things,thing.name,thing)
    def find_thing(self, thing):
        if StringType==type(thing):
            #return found list or empty list
            return self.things.get(thing,[])
        found=[]
        for t in self.things.get(thing.name,[]):
            if t==thing: found.append(t)
        return found
    def remove_thing(self, thing):
        #I don't own this anymore (it may not exist)
        dictlist.remove_value(self.things, thing)
        #remove that 'pointer' from jobs too
        for j in self.current_jobs:
            for (key,value) in j.main_task_info.objects.items():
                if value==thing:
                    j.main_task_info.objects[key]=None
    def find_jobs(self,time):
        #remove stale jobs
        for j in self.current_jobs[:]:
            a=j.main_task
            if time!=a.time: self.current_jobs.remove(j)
        #find new current jobs
        for a in self.tasks:
            current_tasks=map(lambda j:j.main_task,self.current_jobs)
            if time==a.time and a not in current_tasks:
                self.current_jobs.append(job(a))
        #sort according importance and original order
        no=-1 #downward so original order is preserved when reversing
        for j in self.current_jobs: j.no=no; no=no-1
        self.current_jobs.sort() #most important is now last one
        self.current_jobs.reverse()
    def execute_most_important_job(self):
        required_things=self.current_jobs[0].list_required_things()
        for t in required_things:
            g=mgoal.acquire_thing(self,t)
            res=self.check_goal(g)
            if res!=None: return res #action required
            self.current_jobs[0].main_task_info.objects[t]=self.things[t][0]
        #Am I where I should be?
        p=self.current_jobs[0].main_task.place
        if p!=self.place:
            return [event("move",what=self,loc=p)]
        #do that job
        output=self.current_jobs[0].do()
        for e in output[:]:
            if e.command=="make":
                #see similar things
                t_list=self.find_thing(e.what)
                for t in t_list:
                    #is it unfinished
                    if t.status<1.0-const.fzero and t.place==self:
                        #found unfinished thing
                        e.command="change"
                        e.what=t
                        e.amount=0.1
                        return [e]
                if len(t_list)>=30:
                    #I think that I have now enough of these things
                    output.remove(e)
        return output
    #goal section
    def check_goal(self, g, depth=0):
        if g.debug: print "\t"*depth+"GOAL: bef fulfilled:",g.__doc__,g.fulfilled
        if g.fulfilled(self): return
        for sg in g.subgoals:
            if type(sg)==FunctionType or type(sg)==MethodType:
                res=sg(self)
                if g.debug: print "\t"*depth+"GOAL: function:",sg,res
            else:
                if g.debug: print "\t"*depth+"GOAL: bef sg:",sg.__doc__
                res=self.check_goal(sg,depth+1)
                if g.debug: print "\t"*depth+"GOAL: aft sg:",sg.__doc__,res
            if res!=None: return res
    def fulfill_goals(self):
        "see if all goals are fulfilled: if not try to fulfill them"
        for g in self.common_goals:
            res=self.check_goal(g)
            if res: return res
    #event handling: 
    #e is current input event
    #may return new events
    def sight_event(self,e):
        if hasattr(e.what,"command"):
            return self.eyes.analyse_event(e.what)
        else:
            self.mem.remember_place(e.what)
    def destroy_event(self,e):
        self.remove_thing(e.what)
        if e.what==self:
            self.destroy()
            return []
    def say_event(self,e):
        #need something easier that just plain text
        #CHEAT!
        w=e.what
        if e.source!=self and w=="I want to buy axe":
            t=self.things.get('axe')
            if t:
                t=t[0]
                self.remove_thing(t)
                return [event("move", what=t, loc=e.source)]
        if hasattr(w,"is_thing"):
            self.add_thing(w)
            w.place=self.get_rknowledge("xyz",w.xyz)
    def think(self,time,input):
        for e in input:
            output=self.analyse_event(e)
            #empty list is output too
            if output!=None: return output
        output=self.fulfill_goals()
        if output: return output
        self.find_jobs(time)
        output=self.execute_most_important_job()
        return output
    def find_thing_by_id(self, id):
        if self.map.things.has_key(id):
            return self.map.things[id]
        e=self.body.get_make_event(id)
        if hasattr(e.loc,"is_thing"):
            e.loc=self.find_thing_by_id(e.loc.id)
        return self.map.make_thing_from_event(e)
    def run_step(self,time,input):
        """run one time step:
           decide what is most important task and do it"""
        if not self.map.things.has_key(self.id):
            self.map.things[self.id]=self
        if const.use_string_events: input=eval(input)
        events_replace_id_with_thing(self,input)
        self.mem.forget()
        output=self.think(time,input)
        for e in output:
            self.mem.remember_event(e)
            if e.command=="move":
                if type(e.loc)==StringType:
                    e.loc=self.get_knowledge("xyz",e.loc)
        self.debug([str(time),input,output])
        if const.use_string_events: return `output`
        return events_replace_thing_with_id(self,output)
